// NEED Maxlik.ext in /include
/*--------------------------------------------------------------------------
 * maxlik.ox - maximization functions mapped from Gauss to Ox
 *--------------------------------------------------------------------------*/
#ifndef CML_INCLUDED
#define CML_INCLUDED

#pragma array_base(1)

#ifdef GNUDRAW
  #include <packages/gnudraw/gnudraw.h>
#else
  #include <oxdraw.h>
#endif

#include <oxstd.h>
#import <maxsqp>

MLEMeth(const par, const parnames, const nbpar, const T) 
{ 
	decl i, mx, tval, abstval, tprob, astval, tp;
	print("%29s", "Coefficient", "%11s", "Std.Error"); 
	println("%9s", "t-value", "%8s", "t-prob"); 
	for (i = 1; i < sizer(par)+1; ++i)								 
   	{
       	print("%-17s", string(parnames[i]), "%#12.6f", par[i][1]); 
   		print("%#11.6f", par[i][2]); 
    	tval = par[i][3]; 
		abstval = fabs(tval); 
    	tprob = 2 * tailt(abstval, T-sizer(par));  
   		println(abstval <= 1e-4 ? "%9.2f" : abstval >= 1e3 ? "%#9.4g" 
	   		: "%#9.4g", tval, "%8.4f", tprob); 
	} 
}

namespace gauss {

	decl _cml_covpar=0;
	decl _cml_parnames=<>;
	decl _cml_maxiters=-1;
	decl __title=" ";
	decl _cml_algorithm=0;
	decl _cml_gradproc=0;
	decl _cml_hessproc=0;
	decl _cml_hesscov=0;
	decl _cml_xprodcov=0;
/*  decl __output=0; declared in ols, see g2ox.ox*/
	decl _cml_bounds=<>;
	decl _cml_finalhess=0; // MODIFIER
	decl _cml_ineqproc=<>;
	decl _cml_c=<>;
	decl _cml_d=<>;
	decl __weight=1;
/*	decl __row=0; declared in ols, see g2ox.ox*/
	decl _cml_active=<>;
	
	cmlset()
	{
		_cml_covpar=0;
		_cml_parnames=<>;
		_cml_maxiters=-1;
		__title=" ";
		_cml_algorithm=0;
		_cml_gradproc=0;
		_cml_hessproc=0;
		_cml_hesscov=0;
		_cml_xprodcov=0;
		__output=1;
		_cml_bounds=<>;
		_cml_finalhess=0;
		_cml_ineqproc=<>;
		_cml_c=<>;
		_cml_d=<>;
		__weight=1;
		__row=0;
		_cml_active=<>;
	}
	static decl g_numder3func;
	static decl g_analder3func;
	static decl g_amhessfunc;
	static decl g_x;
	g_numder3map(const vP, const adFunc, const avScore, const amHessian)
	{
		adFunc[1] = sumc(__weight.*g_numder3func(vP,g_x));
	    if (avScore)
			avScore[1] = meanc(__weight.*g_analder3func(vP,g_x));
	    if (amHessian)
		{
			amHessian[1] = g_amhessfunc(vP,g_x);
		}
		return 1;
	}
	g_numder3fct(const vP)
	{
		return g_numder3func(vP,g_x);
	}
	static decl g_cfunc;
	g_cfunc_gt0(const avF, const vP)
	{
		decl c=new matrix[rows(g_cfunc(vP))][1];
		c=g_cfunc(vP);
		avF[1]=c;
		return 1;
	}
	static decl g_c;
	static decl g_d;
	g_cd(const avF, const vP)
	{
		avF[1]=-g_d+g_c*vP;
		return 1;
	}
	
	static decl vP;
	static decl m_vGrad;
	static decl m_dLL;
	static decl m_mCovar;
	static decl resp;
	static decl m_cT;
	cmlprt(const proc,...)
	{
		decl out=0;
		if (!ismatrix(proc))
			out={ vP, m_dLL/m_cT, meanc(m_vGrad), m_mCovar, resp };
		return out;
	}
	cml(const x, const name, const f, const vp)
	{
	    decl mhess,hs,asx,cfunc=0;
		decl m_vStdErrors,parnames={};
		decl Tstat,mpar,i,ir,vLo=new matrix[rows(vp)][1],vHi=new matrix[rows(vp)][1];
		vP=vp;
		if (__row!=0)
			println("Warning: option __row <> 0 not supported.");
		if (_cml_bounds==<>)
		{
			vLo[][]=M_INF_NEG;
			vHi[][]=M_INF_POS;
		}
		else
		{
			vLo=_cml_bounds[][1];
			vHi=_cml_bounds[][2];
		}		
		if ((rows(_cml_active)!=<>)&&(rows(_cml_active)==rows(vp)))
		{
			decl fixed=selectifr(range(1,rows(vp))',_cml_active.==0);
			vLo[fixed][]=vp[fixed]; vHi[fixed][]=vp[fixed];
		}
		
		if (isfunction(_cml_ineqproc))
		{
			g_cfunc=_cml_ineqproc;
			cfunc=g_cfunc_gt0;
		}
		if ((_cml_c!=<>)&&(_cml_d!=<>))
		{
			g_d=_cml_d;
			g_c=_cml_c;			
			cfunc=g_cd;
		}
			
	    g_numder3func = f;
	    g_analder3func = _cml_gradproc;
		g_amhessfunc = _cml_hessproc;

		if ((name==0)&&(!isstring(x)))
			g_x=x;
		else if ((!isstring(name))&&(!isstring(x)))
			g_x=x[][name];
		else if ((name==0)&&(isstring(x)))
			[g_x, asx]=__loadd__(x, 2);
		else
		{
			decl all_x, sel_x;
			[all_x, asx]=__loadd__(x, 2);
			decl rr=rows(asx); if (rr==0) ++rr;
			decl rr2=rows(name); if (rr2==0) ++rr2;
			decl asx_double=new matrix[rr][1];
			decl name_double=new matrix[rr2][1];
			for (i=1;i<rr+1;++i)	asx_double[i][]=double(asx[i]);
			for (i=1;i<rr2+1;++i)	name_double[i][]=double(name[i]);
			
			sel_x=indcv(name_double, asx_double);
			g_x=all_x[][sel_x];
		}
		m_cT=rows(g_x);
		MaxControl(_cml_maxiters, __output);
		if (isfunction(g_amhessfunc))
			resp=MaxSQP(g_numder3map, &vP, &m_dLL, &g_amhessfunc, !isfunction(g_analder3func),cfunc, 0, vLo, vHi);
		else
			resp=MaxSQP(g_numder3map, &vP, &m_dLL, 0, !isfunction(g_analder3func),cfunc, 0, vLo, vHi);
		println(string(__title));
		println("Mean log-likelihood: ",m_dLL/m_cT);
		println("Number of cases: ",m_cT);
		println(" ");
		if ((isfunction(g_analder3func))&&(isfunction(_cml_hessproc)))
			println(MaxConvergenceMsg(resp)," using analytical first derivatives and hessian ");
		else if (isfunction(g_analder3func)&&(!isfunction(_cml_hessproc))) 
			println(MaxConvergenceMsg(resp)," using analytical first derivatives ");
		else
			println(MaxConvergenceMsg(resp)," using numerical first derivatives ");
		if (isfunction(_cml_hessproc))
		{
			hs = _cml_hessproc(vP,g_x);
			_cml_hesscov=invert(-hs);
		}
		else
		{
			if (Num2Derivative(g_numder3map, vP, &mhess))
			{
				hs = invert(mhess);
				_cml_hesscov=-hs;
			}
		}
		if (isfunction(g_analder3func))
		   	m_vGrad=g_analder3func(vP,g_x);
		else
			m_vGrad=gradp(g_numder3fct,vP);
		oxwarning(0);
		_cml_xprodcov=invert(m_vGrad'*m_vGrad);
		oxwarning(-1);
		println("Covariance matrix of the parameters computed by the following method:");	
		if (_cml_covpar==0)
		{
			println("Inverse of computed Hessian");
			m_mCovar = _cml_hesscov;  
			if (_cml_hesscov==0)
				println("Warning: singular matrix");
		}
		if (_cml_covpar==1)	// TO BE DONE
		{
			println("Warning Inverse of the second derivatives not implemented (_cml_covpar=1)");
			println("Inverse of computed Hessian instead (_cml_covpar=0)");
			m_mCovar = _cml_hesscov;  
		}
		if (_cml_covpar==2)	
		{
			println("Inverse of the cross-product of the first derivatives");		
			if (_cml_xprodcov==0)
				println("Warning: singular matrix");
			m_mCovar = _cml_xprodcov; 
		}
		if (_cml_covpar==3)
		{
			println("hetereskedastic-consistent covariance matrix");
			m_mCovar = hs*(m_vGrad'*m_vGrad)*hs; 
		}
		m_vStdErrors = sqrt(diagonal(m_mCovar)); 
		Tstat  = vP'./m_vStdErrors; 
	    mpar = vP ~ m_vStdErrors' ~ Tstat'; 
		if (rows(_cml_parnames)!=rows(vP))
		{
			for (i=1; i < rows(vP)+1; ++i)		parnames |= sprint("X(",i,")");
		}
		else
			parnames=_cml_parnames;
		println(" ");			
		MLEMeth(mpar,parnames,rows(vP),m_cT);
		return { vP, m_dLL/m_cT, meanc(m_vGrad), m_mCovar, resp };
	}
	cmlclprt(const b,const f,const g,const h, const ret)
	{
	   __printunsup__("cmlclprt");
	}
	cmltlimits(const b, const h)
	{
	   __printunsup__("cmltlimits");
	}
	cmlclimits(const b, const h)
	{
	   __printunsup__("cmlclimits");
	}
	cmlhist(const b, const h)
	{
	   __printunsup__("cmlhist");
	}
	cmldensity(const b, const h)
	{
	   __printunsup__("cmldensity");
	}
	cmlboot(const x, const name, const f, const vp)
	{
	   __printunsup__("cmlboot");
		return { 0, 0, 0, 0, 0 };
	   
	}
	cmlblimits(const c)
	{
	   __printunsup__("cmlblimits");
	}

	
}