// NEED Maxlik.ext in /include
/*--------------------------------------------------------------------------
 * maxlik.ox - maximization functions mapped from Gauss to Ox
 *--------------------------------------------------------------------------*/
#ifndef MAXLIK_INCLUDED
#define MAXLIK_INCLUDED

#pragma array_base(1)

#include <oxstd.h>

#ifdef GNUDRAW
  #include <packages/gnudraw/gnudraw.h>
#else
  #include <oxdraw.h>
#endif

#include <maximize.h>
#import <maxsqp>

MLEMeth(const par, const parnames, const nbpar, const T) 
{ 
	decl i, mx, tval, abstval, tprob, astval, tp;
	print("%29s", "Coefficient", "%11s", "Std.Error"); 
	println("%9s", "t-value", "%8s", "t-prob"); 
	for (i = 1; i < sizer(par)+1; ++i)								 
   	{
       	print("%-17s", string(parnames[i]), "%#12.6f", par[i][1]); 
   		print("%#11.6f", par[i][2]); 
    	tval = par[i][3]; 
		abstval = fabs(tval); 
    	tprob = 2 * tailt(abstval, T-sizer(par));  
   		println(abstval <= 1e-4 ? "%9.2f" : abstval >= 1e3 ? "%#9.4g" 
	   		: "%#9.4g", tval, "%8.4f", tprob); 
	}
}

namespace gauss {

	decl _max_covpar=0;
	decl _max_parnames=<>;
	decl _max_maxiters=-1;
	decl __title=" ";
	decl _max_algorithm=0;
	decl _max_gradproc=<>;
	decl _max_hessproc=<>;
	decl _max_hesscov=0;
	decl _max_xprodcov=0;
	decl _max_active=<>;
	decl _max_finalhess=0;
	decl __weight=1;

	decl _mlalgr=<>;
	decl _mlgdprc=<>;  
	decl _mlhsprc=<>; 
	decl _mlcovp=<>;
	decl _mlmiter=<>;

	maxset()
	{
		_max_covpar=0;
		_max_parnames=<>;
		_max_maxiters=-1;
		__title=" ";
		_max_algorithm=0;
		_max_gradproc=<>;
		_max_hessproc=<>;
		_max_hesscov=0;
		_max_xprodcov=0;
		_max_finalhess=0;
		__output=1;
		__row=0;
		_max_active=<>;
		__weight=1;
	}
	static decl g_numder3func;
	static decl g_analder3func;
	static decl g_amhessfunc;
	static decl g_x;
	g_numder3map(const vP, const adFunc, const avScore, const amHessian)
	{
		adFunc[1] = sumc(__weight.*g_numder3func(vP,g_x));
	    if (avScore)
			avScore[1] = meanc(__weight.*g_analder3func(vP,g_x));			
	    if (amHessian)
			amHessian[1] = g_amhessfunc(vP,g_x);
		return 1;
	}
	g_numder3fct(const vP)
	{
		return g_numder3func(vP,g_x);
	}
	static decl vP;
	static decl m_vGrad;
	static decl m_dLL;
	static decl m_mCovar;
	static decl resp;
	static decl m_cT;
	maxprt(const proc,...)
	{
		decl out=0;
		if (!ismatrix(proc))
			out={ vP, m_dLL/m_cT, meanc(m_vGrad), m_mCovar, resp };
		return out;
	}
	maxlik(const x, const name, const f, const vp)
	{
		if (isint(_mlalgr))
			_max_algorithm=_mlalgr;
		if (isfunction(_mlgdprc))
			_max_gradproc=_mlgdprc;
		if (isfunction(_mlhsprc))
			_max_hessproc=_mlhsprc;
		if (isint(_mlcovp))
			_max_covpar=_mlcovp;
		if (isint(_mlmiter))
			_max_maxiters=_mlmiter;
			
	    decl mhess,hs,asx;
		decl m_vStdErrors,parnames={};
		decl Tstat,mpar,i,ir;
	    g_numder3func = f;
	    g_analder3func = _max_gradproc;
		g_amhessfunc = _max_hessproc;
		if (__row!=0)
		{
			println("Warning: option __row <> 0 not supported.");
			exit(0);
		}
		if ((name==0)&&(!isstring(x)))
			g_x=x;
		else if ((name==0)&&(isstring(x)))
			[g_x, asx]=__loadd__(x, 2);
		else
		{
			decl all_x, sel_x;
			[all_x, asx]=__loadd__(x, 2);
			decl rr=rows(asx); if (rr==0) ++rr;
			decl rr2=rows(name); if (rr2==0) ++rr2;
			decl asx_double=new matrix[rr][1];
			decl name_double=new matrix[rr2][1];
			for (i=1;i<rr+1;++i)	asx_double[i][]=double(asx[i]);
			for (i=1;i<rr2+1;++i)	name_double[i][]=double(name[i]);
			
			sel_x=indcv(name_double, asx_double);
			g_x=all_x[][sel_x];
		}
		m_cT=rows(g_x);
		MaxControl(_max_maxiters, __output);
//		MaxControlEps(1e-4, 5e-3);
		oxwarning(0);
		vP=vp;
		if (rows(_max_active)==rows(vP))
		{
			decl vLo=new matrix[rows(vp)][1],vHi=new matrix[rows(vp)][1];
			decl fixed=selectifr(range(1,rows(vP))',_max_active.==0);
			vLo[][]=M_INF_NEG;vHi[][]=M_INF_POS;
			vLo[fixed][]=vP[fixed]; vHi[fixed][]=vP[fixed];
			if (isfunction(g_amhessfunc))
				resp=MaxSQPF(g_numder3map, &vP, &m_dLL, &g_amhessfunc, !isfunction(g_amhessfunc),0, 0, vLo, vHi);
			else
				resp=MaxSQPF(g_numder3map, &vP, &m_dLL, 0, !isfunction(g_amhessfunc),0, 0, vLo, vHi);
		}
		else
		{
			if (isfunction(g_amhessfunc))
			{
				MaxControl(20, 0);
				resp=MaxBFGS(g_numder3map, &vP, &m_dLL, 0, !isfunction(g_analder3func));			
				MaxControl(_max_maxiters, __output);
				resp=MaxNewton(g_numder3map, &vP, &m_dLL, &g_amhessfunc, 0);
			}
			else
				resp=MaxBFGS(g_numder3map, &vP, &m_dLL, 0, !isfunction(g_analder3func));			
		}
		oxwarning(-1);
		println(string(__title));
		println("Mean log-likelihood: ",m_dLL/m_cT);
		println("Number of cases: ",m_cT);
		println(" ");
		if ((isfunction(g_analder3func))&&(isfunction(_max_hessproc)))
			println(MaxConvergenceMsg(resp)," using analytical first derivatives and hessian ");
		else if (isfunction(g_analder3func)&&(!isfunction(_max_hessproc))) 
			println(MaxConvergenceMsg(resp)," using analytical first derivatives ");
		else
			println(MaxConvergenceMsg(resp)," using numerical first derivatives ");
		if (isfunction(_max_hessproc))
		{
			hs = _max_hessproc(vP,g_x);
			_max_hesscov=invert(-hs);
		}
		else
		{
			if (Num2Derivative(g_numder3map, vP, &mhess))
			{
				hs = invert(mhess);
				_max_hesscov=-hs;
			}
		}
		_max_finalhess=hs;
		if (isfunction(g_analder3func))
		   	m_vGrad=g_analder3func(vP,g_x);
		else
			m_vGrad=gradp(g_numder3fct,vP);

		_max_xprodcov=invert(m_vGrad'*m_vGrad);
		if (__output) println("Covariance matrix of the parameters computed by the following method:");	
		if (_max_covpar==0)
		{
			if (__output) println("Inverse of computed Hessian");
			m_mCovar = _max_hesscov;  
		}
		if (_max_covpar==1)	// TO BE DONE
		{
			if (__output) println("Warning Inverse of the second derivatives not implemented (_max_covpar=1)");
			println("Inverse of computed Hessian instead (_max_covpar=0)");
			m_mCovar = _max_hesscov;  
		}
		if (_max_covpar==2)	
		{
			if (__output) println("Inverse of the cross-product of the first derivatives");		
			m_mCovar = _max_xprodcov; 
		}
		if (_max_covpar==3)
		{
			if (__output) println("hetereskedastic-consistent covariance matrix");
			m_mCovar = hs*(m_vGrad'*m_vGrad)*hs; 
		}
		m_vStdErrors = sqrt(diagonal(m_mCovar)); 
		Tstat  = vP'./m_vStdErrors; 
	    mpar = vP ~ m_vStdErrors' ~ Tstat'; 
		if (rows(_max_parnames)!=rows(vP))
		{
			for (i=1; i < rows(vP)+1; ++i)		parnames |= sprint("X(",i,")");
		}
		else
			parnames=_max_parnames;
		println(" ");			
		if (__output) MLEMeth(mpar,parnames,rows(vP),m_cT);
		if (rows(_max_active)==rows(vP))
		{
			if (__output)
				println("Warning: constraint optimisation (see _max_active)");
		}
		return { vP, m_dLL/m_cT, meanc(m_vGrad), m_mCovar, resp };
	}
}